
# Copyright (c) 2009 Greg Pintilie - pintilie@mit.edu

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import chimera
import os
import os.path
import Tkinter
from CGLtk import Hybrid
import VolumeData
import _multiscale
import MultiScale.surface
import _surface
import numpy
import _contour
import Matrix
import VolumeViewer
from sys import stderr
from time import clock

from axes import prAxes
from regions import Segmentation, Contact, segmentations
import regions
import graph
from Segger import getMod, dev_menus, timing

OML = chimera.openModels.list

REG_OPACITY = 0.45


def umsg ( txt ) :
    print txt
    status ( txt )

def status ( txt ) :
    txt = txt.rstrip('\n')
    msg.configure(text = txt)
    msg.update_idletasks()

class Volume_Segmentation_Dialog ( chimera.baseDialog.ModelessDialog ):

    title = "Segment Map"
    name = "segment map"
    buttons = ('Segment', 'Group', 'Ungroup', 'Options', "Close")
    help = 'http://ncmi.bcm.edu/ncmi/software/segger/'

    def fillInUI(self, parent):

        self.group_mouse_mode = None

        tw = parent.winfo_toplevel()
        self.toplevel_widget = tw
        tw.withdraw()

        parent.columnconfigure(0, weight = 1)

        row = 1

        menubar = Tkinter.Menu(parent, type = 'menubar', tearoff = False)
        tw.config(menu = menubar)

        file_menu_entries = (
            ('Open segmentation...', self.OpenSegmentation),
            ('Save segmentation', self.SaveSegmentation),
            ('Save segmentation as...', self.SaveSegmentationAs),
            ("Save selected regions to .mrc file...", self.WriteSelRegionsMRCFile),
            ("Save all regions to .mrc file...", self.WriteAllRegionsMRCFile),
            ("Save each region to .mrc file...", self.WriteEachRegionMRCFile),
            ("Close segmentation", self.CloseSeg),
            ("Close all segmenations except displayed", self.CloseHiddenSeg),
            ("Close all segmentations", self.CloseAll),
            )

        fmenu = Hybrid.cascade_menu(menubar, 'File', file_menu_entries)

        regions_menu_entries = (
            'separator',
            ("Show all", self.RegSurfsShowAll),
            ("Show only selected", self.RegSurfsShow),
            ("Show adjacent", self.RegSurfsShowAdjacent),
            ("Hide", self.RegSurfsHide),
            ("Make transparent", self.RegSurfsTransparent),
            ("Make opaque", self.RegSurfsOpaque),
            ('Color density map', self.ColorDensity),
            ("Invert selection", self.RegSelInvert),
            'separator',
            ("Group selected", self.JoinSelRegs),
            ("Ungroup selected", self.UngroupSelRegs),
            ("Smooth and group", self.SmoothAndGroupOneStep),
            ("Delete selected", self.DelSelRegs),
            "separator",
            ("Enclosed volume", self.RegionsVolume),
            ("Mask map with selected", self.MaskMapWRegions),
            ("Show axes for selected", self.ShowRegionAxesSelected),
            ("Hide all axes", self.HideRegionAxes),
            )

        self.regsVisMode = Tkinter.StringVar()
        self.regsVisMode.set ( 'Voxel_Surfaces' )

        rmenu = Hybrid.cascade_menu(menubar, 'Regions', regions_menu_entries)

        if dev_menus:
            rmenu.add_separator()
            for lbl, var, val, cmd in (
                ("Surfaces around voxels", self.regsVisMode,
                 'Voxel_Surfaces', self.RegsDispUpdate),
                ("Map iso-surfaces", self.regsVisMode,
                 'Iso_Surfaces', self.RegsDispUpdate),
                #(" - delete files", self.deleteCloseFiles, 1, None),
                ):
                rmenu.add_radiobutton(label = lbl, variable = var, value = val,
                                      command = cmd)
            for lbl, cmd in (#("Adjacency graph",self.RegionsAdjGraph),
                ('Show group surfaces', self.ShowGroupSurfaces),
                ('Show ungrouped surfaces', self.ShowUngroupedSurfaces),
                ('Select groups', self.SelectGroups),
                ('Select boundary regions', self.SelectBoundaryRegions),
                ("Select non-placed", self.SelectNonPlacedRegions),
                #("Apply threshold", self.RegsDispThr),
                ("Group connected", self.GroupConnectedRegs),
                ("Group by contacts", self.GroupByContacts),
                ("Group using all fits", self.GroupUsingFits),
                #("Ungroup ALL", self.UngroupAllRegs),
                #("Reduce map", self.ReduceMap),
                #("Close", self.CloseRegions),
                ):
                rmenu.add_command(label = lbl, command = cmd)
            #self.deleteCloseFiles = Tkinter.IntVar()
            #self.deleteCloseFiles.set ( 1 )

        if 1 or dev_menus:
            graph_menu_entries = (
                #('Save graph', self.SaveGraph),
                #('Load graph', self.LoadGraph),
                #'separator',
                ('Create graph with uniform link radii', self.Graph),
                (' - Use maximum density between regions', self.GraphMaxD),
                (' - Use average density between regions', self.GraphAvgD),
                (' - Use number of contacting voxel pairs', self.GraphN),
                'separator',
                ('Close graph', self.CloseGraph),
                #('Break selected links', graph.break_selected_links),
                #('Link selected', graph.link_selected),
                #('Group regions using skeleton', self.GroupBySkeleton),
                #('Join selected', self.GroupBySkeleton)
                )
            smenu = Hybrid.cascade_menu(menubar, 'Graph',
                                        graph_menu_entries)

        self.UseAllMods = Tkinter.IntVar()
        self.UseAllMods.set ( 0 )

        from chimera.tkgui import aquaMenuBar
        aquaMenuBar(menubar, parent, row = 0, columnspan=3)


        f = Tkinter.Frame(parent)
        f.grid(column=0, row=row, sticky='ew')
        row += 1

        l = Tkinter.Label(f, text='Segment map')
        l.grid(column=0, row=0, sticky='w')

        self.dmap = Tkinter.StringVar(parent)

        self.mb  = Tkinter.Menubutton ( f, textvariable=self.dmap, relief=Tkinter.RAISED )
        self.mb.grid (column=1, row=0, sticky='we', padx=5)
        self.mb.menu  =  Tkinter.Menu ( self.mb, tearoff=0, postcommand=self.MapMenu )
        self.mb["menu"]  =  self.mb.menu


        f = Tkinter.Frame(parent)
        f.grid(column=0, row=row, sticky='ew')
        row += 1

        l = Tkinter.Label(f, text='Current segmentation')
        l.grid(column=0, row=0, sticky='w')

        self.cur_seg = None
        self.regions_file = Tkinter.StringVar(parent)

        rmb  = Tkinter.Menubutton ( f, textvariable=self.regions_file,
                                    relief=Tkinter.RAISED )
        rmb.grid (column=1, row=0, sticky='we', padx=5)
        rmb.menu  =  Tkinter.Menu ( rmb, tearoff=0,
                                    postcommand=self.FillSegmentationMenu )
        self.mbSegmentationMenu = rmb.menu
        rmb["menu"]  = rmb.menu

        rc = Tkinter.Label(f, text='')
        rc.grid(column=2, row=0, sticky='w')
        self.regionCount = rc


        if dev_menus:
            cp = Hybrid.Popup_Panel(parent)
            cpf = cp.frame
            cpf.grid(row = row, column = 0, sticky = 'news')
            cpf.grid_remove()
            cpf.columnconfigure(0, weight=1)
            self.contactsPanel = cp.panel_shown_variable
            row += 1
            orow = 0

            cb = cp.make_close_button(cpf)
            cb.grid(row = orow, column = 0, sticky = 'e')

            l = Tkinter.Label(cpf, text='Contact Grouping', font = 'TkCaptionFont')
            l.grid(column=0, row=orow, sticky='w', pady=5)
            orow += 1

            s = Hybrid.Scale(cpf, 'Coloring level ', 1, 100, 1, 100)
            s.frame.grid(row = orow, column = 0, sticky = 'ew')
            orow += 1
            self.colorLevel = s
            s.callback(self.SetColorLevel)

            b = Tkinter.Button(cpf, text = 'Set Grouping',
                               command = self.SetContactGrouping)
            b.grid(row = orow, column = 0, sticky = 'w')
            orow += 1

        op = Hybrid.Popup_Panel(parent)
        opf = op.frame
        opf.grid(row = row, column = 0, sticky = 'news')
        opf.grid_remove()
        opf.columnconfigure(0, weight=1)
        self.optionsPanel = op.panel_shown_variable
        row += 1
        orow = 0

        cb = op.make_close_button(opf)
        cb.grid(row = orow, column = 0, sticky = 'e')

        l = Tkinter.Label(opf, text='Segmenting Options', font = 'TkCaptionFont')
        l.grid(column=0, row=orow, sticky='w', pady=5)
        orow += 1

        sopt = Tkinter.Frame(opf)
        sopt.grid(column=0, row=orow, sticky='ew', padx=10)
        orow += 1
        sorow = 0


        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')
        sorow += 1

        l = Tkinter.Label(f, text='Smoothing steps')
        l.grid(column=0, row=0, sticky='w')

        self.numSteps = ns = Tkinter.StringVar(sopt)
        ns.set ( '3' )
        e = Tkinter.Entry(f, width=2, textvariable=self.numSteps)
        e.grid(column=1, row=0, sticky='w', padx=5)

        l = Tkinter.Label(f, text='step size')
        l.grid(column=2, row=0, sticky='w')

        self.stepSize = ss = Tkinter.StringVar(sopt)
        ss.set('1')
        e = Tkinter.Entry(f, width=5, textvariable=self.stepSize)
        e.grid(column=3, row=0, sticky='w', padx=5)

        l = Tkinter.Label(f, text='voxels')
        l.grid(column=4, row=0, sticky='w')


        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')
        sorow += 1

        l = Tkinter.Label(f, text='Keep only regions having at least')
        l.grid(column=0, row=0, sticky='w')

        self.minRegionSize = Tkinter.StringVar(parent)
        self.minRegionSize.set ( '1' )
        e = Tkinter.Entry(f, width=5, textvariable=self.minRegionSize)
        e.grid(column=1, row=0, sticky='w', padx=5)
        e.bind('<KeyPress-Return>', lambda e: self.RemoveSmallRegions())


        l = Tkinter.Label(f, text='voxels')
        l.grid(column=2, row=0, sticky='w')


        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')
        sorow += 1

        l = Tkinter.Label(f, text='Stop grouping if number of regions <=')
        l.grid(column=0, row=0, sticky='w')

        self.targNRegions = Tkinter.StringVar(sopt)
        self.targNRegions.set ( '1' )
        e = Tkinter.Entry(f, width=5, textvariable=self.targNRegions)
        e.grid(column=1, row=0, sticky='w', padx=5)


        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')
        sorow += 1

        l = Tkinter.Label(f, text='Display at most')
        l.grid(column=0, row=0, sticky='w')

        self.maxNumRegions = Tkinter.StringVar(sopt)
        self.maxNumRegions.set ( '2000' )
        e = Tkinter.Entry(f, width=5, textvariable=self.maxNumRegions)
        e.grid(column=1, row=0, sticky='w', padx=5)
        e.bind('<KeyPress-Return>', self.NewMaxRegions)

        l = Tkinter.Label(f, text='regions')
        l.grid(column=2, row=0, sticky='w')


        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')
        sorow += 1

        l = Tkinter.Label(f, text='Surface granularity')
        l.grid(column=0, row=0, sticky='w')

        self.surfaceGranularity = Tkinter.StringVar(sopt)
        self.surfaceGranularity.set ( '1' )
        e = Tkinter.Entry(f, width=5, textvariable=self.surfaceGranularity)
        e.bind('<KeyPress-Return>', self.NewSurfaceResolution)
        e.grid(column=1, row=0, sticky='w', padx=5)

        l = Tkinter.Label(f, text='voxels')
        l.grid(column=2, row=0, sticky='w')


        if dev_menus:
            f = Tkinter.Frame(sopt)
            f.grid(column=0, row=sorow, sticky='w')
            sorow += 1

            l = Tkinter.Label(f, text='Minimum connecting voxels ')
            l.grid(column=0, row=0, sticky='w')

            self.minConnection = Tkinter.StringVar(sopt)
            self.minConnection.set ( '0' )
            e = Tkinter.Entry(f, width=5, textvariable=self.minConnection)
            e.grid(column=1, row=0, sticky='w', padx=5)

        if dev_menus:

            mmf = Tkinter.Frame(sopt)
            mmf.grid(row = sorow, column = 0, sticky = 'ew')
            sorow += 1
    
            mg = Hybrid.Checkbutton(mmf, 'Group with mouse ', False)
            mg.button.grid(row = 0, column = 0, sticky = 'w')
            self.mouse_group = mg.variable
            mg.callback(self.mouse_group_cb)

            mgb = Hybrid.Option_Menu(mmf, '', 'button 1', 'button 2',
                                     'button 3', 'ctrl button 1',
                                     'ctrl button 2', 'ctrl button 3')
            mgb.variable.set('button 3')
            mgb.frame.grid(row = 0, column = 1, sticky = 'w')
            mgb.add_callback(self.mouse_group_button_cb)
            self.mouse_group_button = mgb

        dummyFrame = Tkinter.Frame(parent, relief='groove', borderwidth=1)
        Tkinter.Frame(dummyFrame).pack()
        dummyFrame.grid(row=row,column=0,columnspan=7, pady=7, sticky='we')

        row += 1

        global msg
        msg = Tkinter.Label(parent, width = 60, anchor = 'w', justify = 'left')
        msg.grid(column=0, row=row, sticky='ew')
        self.msg = msg

        umsg ( 'Select an open density map in the field above and press Segment!' )
        row += 1

        vlist = VolumeViewer.volume_list()
        if vlist:
            self.dmap.set ( vlist[0].name )

        for m in segmentations() :

            try : m.seg_map
            except : continue

            if m.display == True :
                self.SetCurrentSegmentation ( m )
                try : self.dmap.set ( m.seg_map.name )
                except : pass

        chimera.openModels.addRemoveHandler(self.ModelClosed, None)

    def SetColorLevel(self):

        smod = self.CurrentSegmentation()
        if smod is None:
            return

        s = self.colorLevel
        lev = int(s.value())

        regions = [r for r in smod.all_regions()
                   if hasattr(r, 'color_level') and r.color_level >= lev and
                   (r.preg is None or r.preg.color_level < lev)]
        smod.color_density(regions)
        return

        # TODO: Unused code adjusts region surface colors.

        if not hasattr(smod, 'contact_grouping'):
            cg = smod.contact_grouping = regions.contact_grouping(smod)
            smod.region_count = len(smod.childless_regions())
        cg = smod.contact_grouping
        range = (smod.region_count - len(cg), smod.region_count)
        if s.range() != range:
            s.set_range(range[0], range[1], step = 1)
        p = max(0, smod.region_count - int(s.value()))

        cs, pairs = regions.connected_subsets(cg[:p])

        # Reset colors
        for r in smod.regions:
            sp = r.surface_piece
            if sp:
                sp.color = r.color

        # Color groups
        for rlist in cs:
            r0 = rlist[0]
            sp0 = r0.surface_piece
            if sp0:
                c = sp0.color
                for r in rlist[1:]:
                    sp = r.surface_piece
                    if sp:
                        sp.color = c

    def SetColorLevelRange(self):

        smod = self.CurrentSegmentation()
        if smod is None:
            return
        
        clevels = [r.color_level for r in smod.all_regions()
                   if hasattr(r, 'color_level')]
        clmin = min(clevels)
        clmax = max(clevels)
        cl = self.colorLevel
        cl.set_range(clmin, clmax, step = 1)
        cl.set_value(clmin)

    def SetContactGrouping(self):

        smod = self.CurrentSegmentation()
        if smod is None:
            return

        s = self.colorLevel
        lev = int(s.value())

        regions = [r for r in smod.all_regions()
                   if hasattr(r, 'color_level') and r.color_level < lev]
        smod.remove_regions(regions)
        self.RegsDispUpdate()


    def ColorDensity(self):

        smod = self.CurrentSegmentation()
        if smod is None:
            return

        smod.color_density()

    def Options(self):

        self.optionsPanel.set(not self.optionsPanel.get())

    def OpenSegmentation(self):

        dmap = getMod ( self.dmap.get() )
        dir = os.path.dirname(dmap.data.path) if dmap else None
        import segfile
        segfile.show_open_dialog(dir, self.OpenSegFiles)

    def OpenSegFiles(self, paths_and_types, open = True):

        from chimera import tasks, CancelOperation
        task = tasks.Task('Loading segmentation', modal = True)
        import segfile
        smods = []
        for path, ftype in paths_and_types:
            if ftype == 'Segmentation':
                try:
                    smod = segfile.read_segmentation(path, open, task)
                except CancelOperation:
                    break
            elif ftype == 'Old regions file':
                dmap = getMod ( self.dmap.get() )
                import regionsfile 
                smod = regionsfile.ReadRegionsFile ( path, dmap )
            smods.append(smod)
            
        if len(smods) == 0:
            return
            
        for smod in smods:
            smod.open_map()
        # TODO: Can't control whether marker model is opened.
        smod = smods[-1]
        self.SetCurrentSegmentation(smod)
        if smod.seg_map:
            self.dmap.set(smod.seg_map.name)
        try:
            self.RegsDispUpdate (task)
        except CancelOperation:
            pass
        task.finished()
        return smods

    def SaveSegmentation(self):

        smod = self.CurrentSegmentation()
        if smod:
            path = smod.path if hasattr(smod, 'path') else None
            import segfile
            segfile.write_segmentation(smod, path)

    def SaveSegmentationAs(self):

        smod = self.CurrentSegmentation()
        if smod:
            import segfile
            segfile.write_segmentation(smod)

    def ModelClosed(self, trigger, n, mlist):

        # Clear menus that are showing closed models.
        for mvar in (self.dmap, self.regions_file):
            if len( mvar.get() ) > 0 and getMod ( mvar.get() ) is None:
                mvar.set('')
        if self.cur_seg in mlist:
            self.cur_seg = None

    def MapMenu ( self ) :

        dmap = None
        if len( self.dmap.get() ) > 0 :
            dmap = getMod ( self.dmap.get() )

        self.mb.menu.delete ( 0, 'end' )        # Clear menu

        open_maps = {}
        
        for m in OML() :
            if type(m) != VolumeViewer.volume.Volume :
                continue

            else :
                # print "adding", m.name
                self.mb.menu.add_radiobutton (
                    label=m.name, variable=self.dmap, command=self.MapSelected )
            


    def MapSelected ( self ) :
        dmap_name = self.dmap.get()
        if len(dmap_name) == 0 : return
        
        dmap = getMod ( dmap_name )
        if dmap == None : return

        dmap.display = True

    def FillSegmentationMenu ( self ) :

        menu = self.mbSegmentationMenu
        menu.delete ( 0, 'end' )      # Clear menu

        sm_map = self.dmap.get()
        smm = getMod ( sm_map )
        if smm == None :
            self.SetCurrentSegmentation(None)
            self.dmap.set ( "" )
            return

        path = os.path.dirname ( smm.data.path ) + os.path.sep
        bname = os.path.splitext ( smm.name ) [0]

        open_names = [(m.name, m) for m in segmentations()]
        if len(open_names ) > 0 :
            menu.add_radiobutton ( label="Open regions files:" )
            menu.add_separator()
            open_names.sort()
            for name, smod in open_names:
                menu.add_radiobutton (label=name, variable=self.regions_file,
                          command=lambda smod=smod: self.RFileSelected(smod) )


        files = os.listdir ( path );
        names_in_path = []
        for f in files :
            if f.find ( bname ) < 0 or f.find('.seg') < 0 : continue
            if f.find ( ".txt" ) >= 0 : continue
            if f.find ( ".mrc" ) >= 0 : continue
            if not f in open_names :
                names_in_path.append ( f )

        if len ( names_in_path ) == 0 : return

        if len(open_names ) > 0 :
            menu.add_separator()

        menu.add_radiobutton ( label="In %s:" % path )
        menu.add_separator()

        for f in names_in_path :
            menu.add_radiobutton (
                label=f, variable=self.regions_file, command=self.RFileSelected )



    def RFileSelected ( self, rmod = None ) :

        if rmod is None:
            mm = getMod ( self.dmap.get() )
            if mm == None : print self.dmap.get(), "not open"; return
            path = os.path.dirname(mm.data.path) + os.path.sep
            rfile = self.regions_file.get()
            rmod = self.OpenSegFiles ( [(path + rfile, 'Segmentation')] )[-1]
        else:
            rmod.display = True
            if rmod.adj_graph : rmod.adj_graph.show_model(True)
            self.ReportRegionCount(rmod)
            for m in segmentations():
                if m != rmod:
                    m.display = False
                    if m.adj_graph : m.adj_graph.show_model(False)
            rfile = rmod.name

        self.cur_seg = rmod

        umsg ( "Showing %s - %d regions, %d surfaces" %
               (rfile, len(rmod.regions), len(rmod.surfacePieces)) )


    def CurrentSegmentation ( self, warn = True ):

        if warn and self.cur_seg is None:
            umsg ( "No segmentation chosen" )
        return self.cur_seg

    def SetCurrentSegmentation ( self, smod ):

        self.cur_seg = smod
        self.regions_file.set(smod.name if smod else '')

    def NewSurfaceResolution( self, event = None ):

        smod = self.CurrentSegmentation()
        if smod is None:
            return

        self.SetSurfaceGranularity(smod)

    def SetSurfaceGranularity ( self, smod ):

        g = self.surfaceGranularity.get()
        try:
            res = float(g)
        except:
            umsg ('Surface granularity "%s" is not a number' % g)
            return

        if res <= 0:
            return

        if smod.regions:
            from chimera import tasks, CancelOperation
            task = tasks.Task('Changing surface resolution', modal = True)
            try:
                smod.change_surface_resolution(res, task)
            except CancelOperation:
                pass
            task.finished()
        else:
            smod.change_surface_resolution(res)


    def NewMaxRegions ( self, event = None ):

        from chimera import tasks, CancelOperation
        task = tasks.Task('Redisplaying regions', modal = True)
        try:
            self.RegsDispUpdate (task)
        except CancelOperation:
            pass
        task.finished()



    def CloseHiddenSeg ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        for m in segmentations () :
            if not m.display:
                umsg ( "Closed %s" % m.name )
                m.close()



    def SaveRegsToMRC ( self, regs, dmap, path = None ) :

        if path is None:
            # Show file chooser dialog.
            fprefix = os.path.splitext(dmap.name)[0]
            if len(regs) == 1 :
                fname = fprefix + "_region_%d.mrc" % regs[0].rid
            else :
                fname = fprefix + "_%d_regions.mrc" % len(regs)
            dir = os.path.dirname ( dmap.data.path )
            import OpenSave
            d = OpenSave.SaveModal ( title = "Save Masked Map",
                                     initialdir = dir, initialfile = fname,
                                     filters = [('MRC map', '*.mrc', '.mrc')] )
            paths_and_types = d.run ( self.toplevel_widget )
            if paths_and_types:
                path = paths_and_types[0][0]
            else:
                return

        (li,lj,lk), (hi,hj,hk) = regions.region_bounds(regs)

        bound = 2
        li = li - bound; lj = lj - bound; lk = lk - bound
        hi = hi + bound; hj = hj + bound; hk = hk + bound

        n1 = hi - li + 1
        n2 = hj - lj + 1
        n3 = hk - lk + 1

        print "Bounds - %d %d %d --> %d %d %d --> %d %d %d" % ( li, lj, lk, hi, hj, hk, n1,n2,n3 )

        umsg ( "Saving %d regions to mrc file..." % len(regs) )

        nmat = numpy.zeros ( (n3,n2,n1), numpy.float32 )
        dmat = dmap.full_matrix()

        #regs_name = ""
        for reg in regs :
            #regs_name = regs_name + "_%d" % reg.rid
            for i,j,k in reg.points() :
                nmat[k-lk,j-lj,i-li] = dmat[k,j,i]
                #for ii in [i-1,i,i+1] :
                #    for jj in [j-1,j,j+1] :
                #        for kk in [k-1,k,k+1] :
                #            nmat[kk-lk,jj-lj,ii-li] = dmat[kk,jj,ii]
                            

        O = dmap.data.origin
        print "origin:", O
        nO = ( O[0] + float(li) * dmap.data.step[0],
               O[1] + float(lj) * dmap.data.step[1],
               O[2] + float(lk) * dmap.data.step[2] )
        
        print "new origin:", nO

        ndata = VolumeData.Array_Grid_Data ( nmat, nO, dmap.data.step, dmap.data.cell_angles )
        try : nv = VolumeViewer.volume.add_data_set ( ndata, None )
        except : nv = VolumeViewer.volume.volume_from_grid_data ( ndata )

        nv.name = os.path.basename ( path )

        nv.openState.xform = dmap.openState.xform

        nv.write_file ( path, "mrc" )

        umsg ( "Wrote %s" % nv.name )



    def WriteSelRegionsMRCFile ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map from which density will be taken" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        smod = self.CurrentSegmentation()
        if smod == None : return

        regs = smod.selected_regions()
        if len(regs)==0 :
            umsg ( "Please select one ore more regions to save to .mrc file" )
            return

        self.SaveRegsToMRC ( regs, dmap )


    def WriteAllRegionsMRCFile ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map from which density will be taken" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        smod = self.CurrentSegmentation()
        if smod == None : return

        regs = [ sp.region for sp in smod.surfacePieces ]

        self.SaveRegsToMRC ( regs, dmap )

        


    def WriteEachRegionMRCFile ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map from which density will be taken" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        smod = self.CurrentSegmentation()
        if smod is None : return

        regs = smod.selected_regions()
        if len(regs)==0 :
            regs = smod.regions

        # Choose file path.
        dir = os.path.dirname ( dmap.data.path )
        fprefix = os.path.splitext ( dmap.name ) [0]
        fname = fprefix + "_region_%d.mrc"
        import OpenSave
        d = OpenSave.SaveModal ( title = "Save Masked Map",
                                 initialdir = dir, initialfile = fname,
                                 filters = [('MRC map', '*.mrc', '.mrc')] )
        paths_and_types = d.run ( self.toplevel_widget )
        if paths_and_types:
            path = paths_and_types[0][0]
        else:
            return
        if not '%d' in path:
            umsg ( "Must include '%d' in map file name for region number" )
            return

        print "Saving each of %d regions to .mrc files" % len(regs)

        for reg in regs :
            self.SaveRegsToMRC ( [reg], dmap, path % (reg.rid,) )



    def MaskMapWRegions ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map from which density will be taken" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        smod = self.CurrentSegmentation()
        if smod is None : return

        regs = smod.selected_regions()
        if len(regs)==0 :
            umsg ( "Please select one ore more regions" )
            return

        dmat = dmap.full_matrix()
        nmat = numpy.zeros_like ( dmat )

        for reg in regs :
            for i,j,k in reg.points() :
                nmat[k,j,i] = dmat[k,j,i]
                            

        ndata = VolumeData.Array_Grid_Data ( nmat, dmap.data.origin, dmap.data.step, dmap.data.cell_angles )
        try : nv = VolumeViewer.volume.add_data_set ( ndata, None )
        except : nv = VolumeViewer.volume.volume_from_grid_data ( ndata )
        nv.name = os.path.splitext ( dmap.name )[0] + "_masked.mrc"
        nv.openState.xform = dmap.openState.xform
        
        nv.surface_levels = [ dmap.surface_levels[0] ]
        nv.update_surface ( False, VolumeViewer.volume.Rendering_Options() )
        for sp in nv.surfacePieces :
            v, t = sp.geometry
            if len(v) == 8 and len(t) == 12 : sp.display = False

        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        nv.write_file ( path + nv.name, "mrc" )
        umsg ( "Wrote %s" % nv.name )

        return nv



    def RegsDispUpdate ( self, task = None ) :

        smod = self.CurrentSegmentation()
        if smod is None : return

        try : smod.seg_map.name
        except : smod.seg_map = getMod ( self.dmap.get() )

        maxnr = self.MaximumRegionsToDisplay()

        if maxnr > 0 and len(smod.regions) >= maxnr  :
            umsg('Only showing %d of %d regions.' % (maxnr, len(smod.regions)))

        smod.display_regions(self.regsVisMode.get(), maxnr, task)

        if maxnr >= len(smod.regions):
            umsg ( "Showing %d region surfaces" % len(smod.regions) )
        else:
            umsg ( "Showing %d of %d region surfaces" %
                   (maxnr, len(smod.regions)) )

        self.ReportRegionCount(smod)

    def MaximumRegionsToDisplay ( self ) :

        try:
            maxnr = int(self.maxNumRegions.get())
        except:
            maxnr = 0
        return maxnr

    def ReportRegionCount ( self, smod ):

        if smod is None:
            s = ''
        else:
            s = "%d regions" %  len(smod.regions)
        self.regionCount["text"] = s


    def RegsDispThr ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        print "%s - thresholding %d regions" % ( smod.name, len(smod.regions) )

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print "Map %s not open" % self.dmap.get(); return
        print " - using map:", dmap.name
        dthr = dmap.surface_levels[0]

        for r in smod.regions : r.remove_surface()

        maxnr = self.MaximumRegionsToDisplay()
        for reg in smod.regions :

            if maxnr > 0 and len(smod.surfacePieces) >= maxnr  :
                umsg('Only showing %d of %d regions.' %
                     (len(smod.surfacePieces), len(smod.regions)))
                break
            reg.make_surface()


    def RegsPrint ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        for reg in smod.regions :

            print "%d - %d %d %d" % ( reg.rid, reg.max_point[0], reg.max_point[1], reg.map_point[2] )


    def ReduceMap ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map first" ); return
        mm = getMod ( self.dmap.get() )
        if mm == None : umsg ( "%s is not open" % self.dmap.get() ); return

        path = os.path.dirname ( mm.data.path ) + os.path.sep
        mname = os.path.splitext ( mm.name )[0]

        d = mm.data
        m2 = d.matrix ( ijk_step=(2,2,2) )
        step2 = ( d.step[0]*2.0, d.step[1]*2.0, d.step[2]*2.0 )
        ld = VolumeData.Array_Grid_Data(m2, d.origin, step2, d.cell_angles, d.rotation,
                           name = mname + '_s2.mrc')

        gv = VolumeViewer.volume.add_data_set ( ld, None )
        gv.name = mname + '_s2.mrc'

        print "writing", path + gv.name
        mod.write_file ( path + gv.name, "mrc" )





    def Segment ( self, show = True, group = True ) :

        from chimera import tasks, CancelOperation
        task = tasks.Task('Segmenting %s' % self.dmap.get(), modal = True)

        try:
            smod = self.SegmentAndGroup(show, group, task)
        except CancelOperation:
            umsg('Cancelled segmentation')
            return
        task.finished()
        return smod


    def SegmentAndGroup ( self, show = True, group = True, task = None ) :

        if len(self.dmap.get()) == 0 : umsg ("Select an open density map to segment first" ); return
        mm = getMod ( self.dmap.get() )
        if mm == None : umsg ( "%s is not open" % self.dmap.get() ); return

        thrD = mm.surface_levels[0]
        print "\n___________________________"
        umsg ( "Segmenting %s, density threshold %f" % (mm.name, thrD) )

        smod = self.CurrentSegmentation(warn = False)
        if smod is None or smod.seg_map != mm:
            mbase, msuf = os.path.splitext ( mm.name )
            msp = msuf.find(' ')
            mend = '' if msp == -1 else msuf[msp:]
            segname = mbase + mend + '.seg'
            smod = Segmentation(segname, mm.openState.xform)
            self.SetSurfaceGranularity(smod)
            self.SetCurrentSegmentation(smod)

        if timing: t0 = clock()
        smod.calculate_local_maxima_regions ( mm, thrD, task )

        if timing: t1 = clock()
        self.RemoveSmallRegions(smod, task)
        nwr = len(smod.regions)

        if timing: t2 = clock()
        if group:
            self.SmoothAndGroup ( smod, task )

        # Undisplay other segmentations
        if timing: t3 = clock()
        for m in segmentations() :
            if m != smod:
                m.display = False

        self.RegsDispUpdate ( task )	 # Display region surfaces
#        mm.display = False              # Undisplay map

        if timing: 
            t4 = clock()
            print "Time %.2f sec: watershed %.2f sec, small %.2f, group %.2f sec, display %.2f sec" % (t4-t0, t1-t0, t2-t1, t3-t2, t4-t3)

        umsg('%d watershed regions, smoothed %d voxels to get %d regions' %
             (nwr, smod.smoothing_level, len(smod.regions)))

        return smod

    def RemoveSmallRegions(self, smod = None, task = None):

        if smod is None:
            smod = self.CurrentSegmentation()
            if smod is None:
                return

        mrs = self.minRegionSize.get()
        try:
            minsize = int ( mrs )
        except:
            print 'Minimum region size "%s" is not an integer' % mrs
            minsize = 1

        if minsize <= 1:
            return

        if task is None:
            from chimera import tasks, CancelOperation
            task = tasks.Task('Removing small regions', modal = True)
            try:
                smod.remove_small_regions(minsize, task)
                self.RegsDispUpdate(task)
            except CancelOperation:
                umsg('Cancelled removing small regions')
                return
            task.finished()
        else:
            smod.remove_small_regions(minsize, task)
            self.RegsDispUpdate(task)

        self.ReportRegionCount(smod)


    def GroupConnectedRegs ( self ) :

        min_contact = int(self.minConnection.get())

        smod = self.CurrentSegmentation()
        if smod == None : return

        regions = smod.selected_regions()
        if len(regions)==0 :
            regions = smod.regions
        
        smod.group_connected ( regions, min_contact)
        self.RegsDispUpdate()
        
        self.ReportRegionCount(smod)

        if regions:
            from regions import TopParentRegions
            nsurfs = [r.surface_piece for r in TopParentRegions(regions)]
            chimera.selection.setCurrent(nsurfs)


    def GroupByContacts ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        from chimera import tasks, CancelOperation
        task = tasks.Task('Contact grouping', modal = True)
        try:
            regions.group_by_contacts(smod, task)
            self.RegsDispUpdate(task)
        except CancelOperation:
            umsg('Cancelled contact grouping')
        task.finished()
            
        self.SetColorLevelRange()
        self.ReportRegionCount(smod)

        self.contactsPanel.set(True)

    def CloseAll ( self ) :

        umsg ( "Closing all segmentations." )
        
        dmap = getMod ( self.dmap.get() )
        if dmap != None :
            dmap.display = True

        for m in segmentations ():
            print 'Closed', m.name
            m.close()

        self.SetCurrentSegmentation ( None )
        self.ReportRegionCount(None)


    def SmoothAndGroup ( self, smod, task = None ) :

        try :
            numit = int ( self.numSteps.get() )
            sdev = float ( self.stepSize.get() )
        except :
            umsg ( "Enter <int> for # smoothing steps, <float> for step size" )
            return

        try : targNRegs = int(self.targNRegions.get())
        except : umsg ( "Enter an integer for target # of regions" ); return

        if targNRegs <= 0 :
            umsg ( "Enter an integer > 0 for target # of regions" )
            return

        smod.smooth_and_group(numit, sdev, targNRegs, task)

        self.ReportRegionCount(smod)



    def SmoothAndGroupOneStep ( self ) :

        smod = self.CurrentSegmentation()
        if smod is None:
            return

        if len(smod.regions) <= 1:
            umsg ('%s has %d regions' % (smod.name, len(smod.regions)))
            return

        try :
            step = float ( self.stepSize.get() )
        except :
            umsg ( "Enter <float> for step size" )
            return
        sdev = step + smod.smoothing_level

        umsg ( "Smoothing and grouping, standard deviation %.3g voxels" % sdev)

        from chimera import tasks, CancelOperation
        task = tasks.Task('Smooth and group', modal = True)
        try:
            while len(smod.smooth_and_group(1, sdev, 1, task)) == 0:
                umsg ('No new groups smoothing %.3g voxels' % sdev)
                sdev += step
            self.RegsDispUpdate ( task )	 # Display region surfaces
        except CancelOperation:
            umsg('Cancelled smooth and group')
            return
        task.finished()

        self.ReportRegionCount(smod)

        if smod.adj_graph :
            graph.create_graph ( smod, smod.graph_links )

        umsg ( "Got %d regions after smoothing %.3g voxels." %
               (len(smod.regions), sdev) )


    def GroupUsingFits ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print "Map %s not open" % self.dmap.get(); return

        smod = self.CurrentSegmentation()
        if smod == None : return

        if len(smod.regions) == 0 : print "No regions in", smod.name; return
        try : dmap.fitted_mols
        except : dmap.fitted_mols = []
        if len(dmap.fitted_mols) == 0 : print "No fits found for", dmap.name; return

        print "Grouping %d regions by overlap to %d fitted structures" % (
            len(smod.regions), len(dmap.fitted_mols) )

        dmap.chain_maps = []

        for mol in dmap.fitted_mols :
            try : mol.fmap.imap
            except : mol.fmap.imap = self.MapIndexesInMap ( dmap, mol.fmap )
            from random import random as rand
            mol.fmap.surf_color = ( rand(), rand(), rand(), 1 )
            dmap.chain_maps.append ( mol.fmap )

#        self.SegAccuracy ( "_fits_acc", True )



    def RegSurfsShowAll ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return
        
        for sp in smod.surfacePieces :
            sp.display = True


    def RegSurfsShow ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        for p in smod.surfacePieces : p.display = False
        for r in smod.selected_regions() :
            r.surface_piece.display = True


    def RegSurfsHide ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        sregs = smod.selected_regions()        
        if len(sregs) == 0 : sregs = smod.regions

        for r in sregs : r.surface_piece.display = False


    def RegSurfsShowAdjacent ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        sregs = smod.selected_regions()        
        if len(sregs) == 0 : sregs = smod.regions

        cr = set()
        for r in sregs :
            cr.update(r.contacting_regions())

        umsg ( "Region has %d adjacent regions" % len(cr) )
        for r in cr :
            r.surface_piece.display = True



    def RegSurfsTransparent ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        sregs = smod.selected_regions()        
        if len(sregs) == 0 : sregs = smod.regions

        for r in sregs :
            clr = r.color
            r.surface_piece.color = ( clr[0], clr[1], clr[2], REG_OPACITY )

        # smod.piecesAreSelectable = False



    def RegSurfsOpaque ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        sregs = smod.selected_regions()        
        if len(sregs) == 0 : sregs = smod.regions

        for r in sregs :
            clr = r.color
            r.surface_piece.color = ( clr[0], clr[1], clr[2], 1.0 )



    def RegSelInvert ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        sel_regs = set ( smod.selected_regions() )
        surfs = [r.surface_piece for r in smod.regions
                 if not r in sel_regs]
                
        chimera.selection.clearCurrent ()
        chimera.selection.addCurrent ( surfs )
    

    def Group ( self ):

        if NothingSelected():
            self.SmoothAndGroupOneStep()
        else:
            self.JoinSelRegs()

    def JoinSelRegs ( self ) :

        smod = self.CurrentSegmentation()
        if smod is None : return

        regs = smod.selected_regions()
        if len(regs)==0 :
            umsg ( "No regions selected" )
            return

        jreg = smod.join_regions ( regs )
        jreg.make_surface(None, None, smod.regions_scale)

        if smod.adj_graph :
            graph.create_graph ( smod, smod.graph_links )

        chimera.selection.setCurrent([jreg.surface_piece])

        self.ReportRegionCount(smod)
        umsg ( "Grouped %d regions" % len(regs) )




    def DelSelRegs ( self ) :

        smod = self.CurrentSegmentation()
        if smod is None : return

        regs = smod.selected_regions()
        if len(regs)==0 :
            umsg ( "Please select one ore more regions to register the structure with" )
            return

        smod.remove_regions ( regs, remove_children = True )

        self.ReportRegionCount(smod)
        umsg ( "Deleted %d regions" % len(regs) )




    def Ungroup ( self ):

        if NothingSelected():
            self.UngroupLastSmoothing()
        else:
            self.UngroupSelRegs()


    def UngroupSelRegs ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        sregs = smod.selected_regions()

        if len(sregs) == 0 :
            umsg ( "No regions selected" )
            return

        chimera.selection.clearCurrent ()

        rlist = smod.ungroup_regions ( sregs )

        for reg in rlist :
            reg.make_surface(None, None, smod.regions_scale)

        if smod.adj_graph :
            graph.create_graph ( smod, smod.graph_links )

        chimera.selection.setCurrent ( [r.surface_piece for r in rlist] )

        self.ReportRegionCount(smod)



    def UngroupAllRegs ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        rlist = list(smod.regions)
        smod.ungroup_regions(rlist)

        umsg ( "Ungrouped %d regions -> %d regions" % (len(rlist), len(smod.regions)) )
        self.ReportRegionCount(smod)



    def UngroupLastSmoothing ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        levels = [r.smoothing_level for r in smod.regions]
        if len(levels) == 0:
            return
        slev = max(levels)

        rlev = [r for r in smod.regions if r.smoothing_level == slev]
            
        from chimera import tasks, CancelOperation
        task = tasks.Task('Ungrouping', modal = True)
        try:
            smod.ungroup_regions(rlev, task)
        except CancelOperation:
            pass
        task.finished()

        levels = [r.smoothing_level for r in smod.regions]
        smod.smoothing_level = max(levels)

        if smod.adj_graph :
            graph.create_graph ( smod, smod.graph_links )

        umsg ( "Ungrouped to %.3g voxel smoothing, %d regions" %
               (smod.smoothing_level, len(smod.regions)) )
        self.ReportRegionCount(smod)


    def CloseRegions ( self ) :

        smod = self.CurrentSegmentation()
        if smod is None : return

        chimera.openModels.remove ( smod )
        self.SetCurrentSegmentation(None)
        self.ReportRegionCount(None)

        if smod.adj_graph : smod.adj_graph.close()


    def CloseSeg ( self ) :

        smod = self.CurrentSegmentation()
        if smod is None : return

        smod.close()
        self.SetCurrentSegmentation(None)



    def RegionsVolume ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        sregs = smod.selected_regions()
        print "%d selected regions" % len(sregs)

        if len(sregs) == 0 :
            sregs = smod.regions

        if len(sregs) == 0 :
            umsg ( "No regions found in %s" % smod.name )
            return

        tvol = sum([reg.enclosed_volume() for reg in sregs])
        pcount = sum([reg.point_count() for reg in sregs])
        
        rw = "region"
        if len(sregs) > 1 : rw = "regions"
        umsg ( "Volume of %d %s: %.3g Angstroms^3, %d points" % ( len(sregs), rw, tvol, pcount ) )



    def Graph ( self ) :

        smod = self.CurrentSegmentation()
        if smod:
            graph.create_graph(smod,"uniform")

    def GraphAvgD ( self ) :

        smod = self.CurrentSegmentation()
        if smod:
            graph.create_graph(smod,"avgd")

    def GraphMaxD ( self ) :

        smod = self.CurrentSegmentation()
        if smod:
            graph.create_graph(smod,"maxd")

    def GraphN ( self ) :

        smod = self.CurrentSegmentation()
        if smod:
            graph.create_graph(smod,"N")


    def LoadGraph ( self ) :

        smod = self.CurrentSegmentation()
        if smod is None:
            graph.open_skeleton(smod)

    def SaveGraph ( self ) :

        smod = self.CurrentSegmentation()
        if smod:
            graph.read_graph(smod)

    def CloseGraph ( self ) :

        smod = self.CurrentSegmentation()
        if smod:
            graph.close(smod)
            smod.display = True

    def GroupBySkeleton ( self ) :

        smod = self.CurrentSegmentation()
        if smod:
            skeleton.group_by_skeleton(smod)
            smod.display = True
            smod.display_regions()
            self.ReportRegionCount(smod)

    def RemoveGraphLinks ( self ) :

        graph.remove_graph_links()



    def ShowRegionsAxes ( self, regs ) :

        smod = self.CurrentSegmentation()
        if smod is None: return

        for r in regs :

            sp = r.surface_piece
            try :
                sp.axes.display = True
                chimera.openModels.close ( sp.axes )
            except :
                pass

            tpoints = r.map_points()
            sp.COM, sp.U, sp.S, sp.V = prAxes ( tpoints )

            com = numpy.sum(tpoints, axis=0) / len(tpoints)
            comv = numpy.ones_like ( tpoints ) * com
            points = tpoints - comv

            ppoints = points * sp.U
            sp.Extents = numpy.asarray ( numpy.max ( numpy.abs ( ppoints ), 0 ) )[0]

            from random import random as rand
            if 0 and rand() > 0.5 :
                sp.U[0,0] = sp.U[0,0] * -1.0
                sp.U[1,0] = sp.U[1,0] * -1.0
                sp.U[2,0] = sp.U[2,0] * -1.0
            if rand() > 0.5 : r_y = -1.0
            if rand() > 0.5 : r_z = -1.0

            sp.Extents[0] = sp.Extents[0] + 5.0
            sp.Extents[1] = sp.Extents[1] + 5.0
            sp.Extents[2] = sp.Extents[2] + 5.0

            import axes
            sp.axes = axes.AxesMod ( sp.COM, sp.U, sp.Extents, 1.0, 1.1,
                                     alignTo = sp.model )
            sp.axes.name = "region_%d_axes" % r.rid



    def ShowRegionAxesSelected ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        sregs = smod.selected_regions()        
        if len(sregs)==0 : print "no selected regions found"; return

        self.ShowRegionsAxes ( sregs )


    def HideRegionAxes ( self ) :

        print "hiding axes"
        for m in OML() :
            t = m.name.split ("_")
            if t[0] == "region" and t[2] == "axes" :
                print "- removing", m.name
                chimera.openModels.close( m )


    def MapIndexesInMap ( self, ref_map, mask_map ) :

        thr = mask_map.surface_levels[0]
        mm = mask_map.data.matrix()
        mm = numpy.where ( mm > thr, mm, numpy.zeros_like(mm) )

        nze = numpy.nonzero ( mm )
        nzs = numpy.array ( [nze[2], nze[1], nze[0]] )

        # the copy is needed! otherwise the _contour.afine_transform does not work for some reason
        points =  numpy.transpose ( nzs ).astype(numpy.float32)

        #points = numpy.zeros ( ( len(nze[0]), 3 ), numpy.float32 )
        #for ei, i in enumerate ( nze[0] ) :
        #    j = nze[1][ei]
        #    k = nze[2][ei]
        #    points[ei][0], points[ei][1], points[ei][2] = float (k), float(j), float(i)

        #print points[0]

        print "Making map indices for %s in %s" % ( mask_map.name, ref_map.name )
        print " - %d points above %.3f" % ( len(points), thr )

        # transform to index reference frame of ref_map
        f1 = mask_map.data.ijk_to_xyz_transform
        f2 = Matrix.xform_matrix ( mask_map.openState.xform )
        f3 = Matrix.xform_matrix ( ref_map.openState.xform.inverse() )
        f4 = ref_map.data.xyz_to_ijk_transform

        tf = Matrix.multiply_matrices( f2, f1 )
        tf = Matrix.multiply_matrices( f3, tf )
        tf = Matrix.multiply_matrices( f4, tf )
        _contour.affine_transform_vertices ( points, tf )

        #_contour.affine_transform_vertices ( points, f1 )
        #_contour.affine_transform_vertices ( points, f2 )
        #_contour.affine_transform_vertices ( points, f3 )

        #print points[0]

        #com = numpy.sum (points, axis=0) / len(points)
        #C = chimera.Vector ( com[0], com[1], com[2] )
        #comv = numpy.ones_like ( points ) * com
        #points_v = points - comv
        #bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points_v), 1 ) ) )

        # transform points to indexes in reference map
        # _contour.affine_transform_vertices ( points, ref_map.data.xyz_to_ijk_transform )

        imap = {}
        for fk, fj, fi in points :

            for i in [ int(numpy.floor(fi)), int(numpy.ceil(fi)) ] :
                for j in [ int(numpy.floor(fj)), int(numpy.ceil(fj)) ] :
                    for k in [ int(numpy.floor(fk)), int(numpy.ceil(fk)) ] :

                        try : mi = imap[i]
                        except : mi = {}; imap[i] = mi
                        try : mij = mi[j]
                        except : mij = {}; mi[j] = mij
                        mij[k] = 1


        return imap #, C, bRad


    def ShowGroupSurfaces ( self ) :
                                         
        smod = self.CurrentSegmentation()
        if smod is None : return

        regs = set([r.top_parent() for r in smod.selected_regions()])
        if len(regs)==0 :
            regs = smod.regions

        for r in regs:
            for c in r.all_children():
                c.remove_surface()
            r.make_surface()


    def ShowUngroupedSurfaces ( self ) :
                                         
        smod = self.CurrentSegmentation()
        if smod is None : return

        regs = set([r.top_parent() for r in smod.selected_regions()])
        if len(regs)==0 :
            regs = smod.regions

        from chimera import tasks, CancelOperation
        task = tasks.Task('Showing ungrouped surfaces', modal = True)
        try:
            for i, r in enumerate(regs):
                if task and i % 100 == 0:
                    task.updateStatus('region %d of %d' % (i, len(regs)))
                if r.has_children():
                    r.remove_surface()
                    for c in r.childless_regions():
                        c.make_surface()
                else:
                    r.make_surface()
        except CancelOperation:
            pass
        task.finished()


    def SelectGroups ( self ) :
                                         
        smod = self.CurrentSegmentation()
        if smod is None : return

        rlist = [r for r in smod.regions if r.has_children()]
        from regions import all_regions
        surfs = [r.surface() for r in all_regions(rlist) if r.has_surface()]
        from chimera import selection
        selection.setCurrent(surfs)



    # Regions that have voxels on the mask boundary.
    def SelectBoundaryRegions ( self, pad = 3 ) :
                                         
        smod = self.CurrentSegmentation()
        if smod is None : return

        m = smod.mask
        if m is None:
            return

        import _segment
        b = _segment.region_bounds(m)

        rset = set()
        kmax, jmax, imax = [(s-1)-pad for s in m.shape]
        for r in smod.childless_regions():
            i = r.rid
            if (i < len(b) and
                (b[i,0] <= pad or b[i,1] <= pad or b[i,2] <= pad or
                 b[i,3] >= kmax or b[i,4] >= jmax or b[i,5] >= imax)):
                rset.add(r.top_parent())
        from regions import all_regions
        rlist = all_regions(rset)
        surfs = [r.surface() for r in rlist if r.has_surface()]
        from chimera import selection
        selection.setCurrent(surfs)


    def SelectNonPlacedRegions ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a density map"); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return
                                         
        smod = self.CurrentSegmentation()
        if smod is None : return

        nsel = 0
        chimera.selection.clearCurrent ()
        for sp in smod.surfacePieces :
            try :
                sp.region.placed
            except :
                chimera.selection.addCurrent ( sp )
                nsel = nsel + 1

        print "%d non-placed regions" % nsel


    def mouse_group_cb(self):

        gmm = self.group_mouse_mode
        if self.mouse_group.get():
            if gmm is None:
                import mousemode
                gmm = mousemode.Group_Connected_Mouse_Mode()
                self.group_mouse_mode = gmm
            button, modifiers = self.mouse_button_spec()
            gmm.bind_mouse_button(button, modifiers)
        elif gmm:
            gmm.unbind_mouse_button()
    
    def mouse_group_button_cb(self):

        if self.mouse_group.get() and self.group_mouse_mode:
            button, modifiers = self.mouse_button_spec()
            self.group_mouse_mode.bind_mouse_button(button, modifiers)
      
    def mouse_button_spec(self):

        name = self.mouse_group_button.variable.get()
        name_to_bspec = {'button 1':('1', []), 'ctrl button 1':('1', ['Ctrl']),
                         'button 2':('2', []), 'ctrl button 2':('2', ['Ctrl']),
                         'button 3':('3', []), 'ctrl button 3':('3', ['Ctrl'])}
        bspec = name_to_bspec[name]
        return bspec


def NothingSelected():

    from chimera import selection
    return selection.currentEmpty()


def volume_segmentation_dialog ( create=False ) :

  from chimera import dialogs
  return dialogs.find ( Volume_Segmentation_Dialog.name, create=create )


def show_volume_segmentation_dialog ():

    from chimera import dialogs
    d = volume_segmentation_dialog ( create = True )
    # Avoid transient dialog resizing when created and mapped for first time.
    d.toplevel_widget.update_idletasks ()
    d.enter()

    from Accelerators import add_accelerator
    add_accelerator('gg', 'Group regions', d.JoinSelRegs )
    add_accelerator('uu', 'Ungroup regions', d.UngroupSelRegs )
    add_accelerator('rh', 'Hide regions', d.RegSurfsHide )
    add_accelerator('rs', 'Hide regions', d.RegSurfsShow )
    add_accelerator('dd', 'Delete regions', d.DelSelRegs )
    print "gg - groups regions"
    print "uu - ungroup regions"

    return d


def current_segmentation(warn = True):

    d = volume_segmentation_dialog()
    if d:
        return d.CurrentSegmentation(warn)
    elif warn:
        umsg ( "No segmentation opened" )
    return None

def segmentation_map():

    d = volume_segmentation_dialog()
    if d:
        dmap_name = d.dmap.get()
        if len(dmap_name) == 0 : return None
        return getMod ( dmap_name )
    return None

# -----------------------------------------------------------------------------
#
from chimera import dialogs
dialogs.register (Volume_Segmentation_Dialog.name, Volume_Segmentation_Dialog,
                  replace = True)

